/* this file is a part of amp software, (C) tomislav uzelac 1996,1997
*/

/* audio.c	main amp source file 
 *
 * Created by: tomislav uzelac	Apr 1996 
 * Karl Anders Oygard added the IRIX code, 10 Mar 1997.
 * Ilkka Karvinen fixed /dev/dsp initialization, 11 Mar 1997.
 * Lutz Vieweg added the HP/UX code, 14 Mar 1997.
 * Dan Nelson added FreeBSD modifications, 23 Mar 1997.
 * Andrew Richards complete reorganisation, new features, 25 Mar 1997
 * Edouard Lafargue added sajber jukebox support, 12 May 1997
 * Ilyes Gouta made a Win32 port of amp.
 */

#pragma optimize("gsy",on)
#pragma comment(linker,"/RELEASE")
#pragma comment(linker,"/opt:nowin98")

#include <windows.h>
#include "amp.h"
#include "gui.h"

#include <stdio.h>
#include <string.h>

#define AUDIO
#include "audio.h"
#include "formats.h"
#include "getbits.h"
#include "huffman.h"
#include "layer2.h"
#include "layer3.h"
#include "position.h"
#include "transform.h"
#include "audioIO.h"

char App[]="G-amp";

static int holdDecoder;
int decodedFrames;

void statusDisplay(struct AUDIO_HEADER *header, int frameNo)
{
	int minutes,seconds;

	if ((A_SHOW_CNT || A_SHOW_TIME) && !(frameNo%10))
		msg("\r");
	if (A_SHOW_CNT && !(frameNo%10) ) {
		msg("{ %d } ",frameNo);
	}
	if (A_SHOW_TIME && !(frameNo%10)) {
		seconds=frameNo*1152/t_sampling_frequency[header->ID][header->sampling_frequency];
		minutes=seconds/60;
		seconds=seconds % 60;
		msg("[%d:%02d]",minutes,seconds);
	}
	if (A_SHOW_CNT || A_SHOW_TIME)
		fflush(stderr);
}

void stopMPEG()
{
	holdDecoder=1;
}

int decodeMPEG(void)
{
	int cnt=0,g,snd_eof;
	struct AUDIO_HEADER header;

	initialise_globals();
	holdDecoder=0;

	if (A_FORMAT_WAVE) wav_begin();
	
	if ((g=gethdr(&header))!=0) {
		report_header_error(g);
		return -1;
	}

	if (header.protection_bit==0) getcrc();

	if (setup_audio(&header)!=0) {
		warn("Cannot set up audio. Exiting\n");
		return -1;
	}
	
	show_header(&header);

	if (header.layer==1) {
		if (layer3_frame(&header,cnt)) {
			warn(" error. blip.\n");
			return -1;
		}
	} else if (header.layer==2)
		if (layer2_frame(&header,cnt)) {
			warn(" error. blip.\n");
			return -1;
		}

	snd_eof=0;
	cnt=0;
	while (!snd_eof&&!holdDecoder/*&& ready_audio()*/) {
		if ((g=gethdr(&header))) {
			report_header_error(g);
			if (g==GETHDR_EOF && A_FORMAT_WAVE) wav_end(&header);
			snd_eof=1;
			break;
		}

		if (header.protection_bit==0) getcrc();

		statusDisplay(&header,cnt);	

		if (header.layer==1) {
			if (layer3_frame(&header,cnt)) {
				warn(" error. blip.\n");
				return -1;
			}
		} else if (header.layer==2) 
			if (layer2_frame(&header,cnt)) {
				warn(" error. blip.\n");
				return -1;
			}
		decodedFrames=++cnt;
	}
	return 0;
}

void initialise_decoder(void)
{
	premultiply();
	imdct_init();
	calculate_t43();
}

void initialise_globals(void)
{
	append=data=nch=0;
	f_bdirty=TRUE;
	bclean_bytes=0;

	ZeroMemory(s,sizeof(s));
	ZeroMemory(res,sizeof(res));
}

void report_header_error(int err)
{
	switch (err) {
		case GETHDR_ERR: die("error reading mpeg bitstream. exiting.\n");
					break;
		case GETHDR_NS : warn("this is a file in MPEG 2.5 format, which is not defined\n");
				 warn("by ISO/MPEG. It is \"a special Fraunhofer format\".\n");
				 warn("amp does not support this format. sorry.\n");
					break;
		case GETHDR_FL1: warn("ISO/MPEG layer 1 is not supported by amp (yet).\n");
					break;
		case GETHDR_FF : warn("free format bitstreams are not supported. sorry.\n");
					break;	
		case GETHDR_SYN: warn("oops, we're out of sync.\n");
					break;
		case GETHDR_EOF: /* need to clean up stuff here */
		default:;
	}	
}

int setup_audio(struct AUDIO_HEADER *header)
{
	if (A_AUDIO_PLAY) audioBufferOpen(t_sampling_frequency[header->ID][header->sampling_frequency],(header->mode!=3 && !A_DOWNMIX),A_SET_VOLUME);
	return 0;
}

void close_audio(void)
{
	if (A_AUDIO_PLAY) audioBufferClose();
}

int ready_audio(void)
{
	return 1;
}

int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR comLine,int show)
{
	int argPos;
	MSG msg;

	argPos=args(0,NULL); //argPos=args(argc,argv); // process command line arguments
	initialise_decoder();	/* initialise decoder */

	init_gui();
	while (GetMessage(&msg,NULL,0,0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return 0;
}