/*
 * Author : Ilyes Gouta
 * Subject : audio driver for amp (Win32)
 * Notes :
 *
 */

#include <windows.h>
#include <stdio.h>
#include <dsound.h>
#include "audio.h"
#include "audioIO.h"

extern "C" char App[];

#define SAFE_RELEASE(a) if (a) {a->Release();a=NULL;}

#define BUFFER_LENGTH                        2000 /* 2000 ms */

#define BUFFER_SIZE                          ((WriteOffset + BufferSize - ReadOffset) % BufferSize)

static LPDIRECTSOUND8 pDS = NULL;
static LPDIRECTSOUNDBUFFER pSoundBuffer = NULL;
static LPDIRECTSOUNDNOTIFY8 pDSNotify = NULL;

static DWORD BufferSize;
static DWORD WriteOffset,ReadOffset;
static int Volume=127;

static HANDLE hAccessNotification, hWritingNotification;
static HANDLE hCtrlThread;
static int PlayBuffer, FirstDecode;
static int RequestedInput;

DWORD WINAPI DSCtrlThread(LPVOID pData)
{
	while (PlayBuffer) {
		WaitForSingleObject(hAccessNotification, INFINITE);
		if (!PlayBuffer) break;

loop:
		pSoundBuffer->GetCurrentPosition(&ReadOffset, NULL);
		if ((BUFFER_SIZE + RequestedInput) < BufferSize) SetEvent(hWritingNotification);
		else {
			Sleep(3); /* will deschedule !! */
			goto loop;
		}

	}
	return 0;
}

int audioOpen(int freq, int stereo, int vol)
{
	DSBUFFERDESC bufferdesc;
	WAVEFORMATEX format;
	DWORD tID;

	ZeroMemory(&format, sizeof(format));
	format.cbSize = sizeof(format);
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = (stereo ? 2 : 1);
	format.nSamplesPerSec = freq;
	format.nBlockAlign = format.nChannels * 2;
	format.wBitsPerSample = 16;
	format.nAvgBytesPerSec = format.nBlockAlign * format.nSamplesPerSec;

	ZeroMemory(&bufferdesc, sizeof(bufferdesc));
	bufferdesc.dwSize = sizeof(bufferdesc);
	bufferdesc.dwFlags = DSBCAPS_CTRLVOLUME | DSBCAPS_GLOBALFOCUS | DSBCAPS_GETCURRENTPOSITION2;
	bufferdesc.dwBufferBytes = (int)(((float)BUFFER_LENGTH/1000) * format.nAvgBytesPerSec);
	bufferdesc.dwBufferBytes -= (bufferdesc.dwBufferBytes % format.nBlockAlign);
	bufferdesc.lpwfxFormat = &format;
	bufferdesc.guid3DAlgorithm = GUID_NULL;

	BufferSize = bufferdesc.dwBufferBytes;
	
	if (DirectSoundCreate8(NULL, &pDS, NULL)!=DS_OK) return 0;
	if (pDS->SetCooperativeLevel(GetDesktopWindow(), DSSCL_NORMAL) != DS_OK) {SAFE_RELEASE(pDS);return 0;}
	if (pDS->CreateSoundBuffer(&bufferdesc, &pSoundBuffer, NULL) != DS_OK) {SAFE_RELEASE(pDS);return 0;}

	void *bytes1,*bytes2;
	DWORD size1,size2;

	pSoundBuffer->Lock(0 ,0 , &bytes1, &size1, &bytes2, &size2, DSBLOCK_ENTIREBUFFER);
	ZeroMemory((BYTE*)bytes1, size1);
	pSoundBuffer->Unlock(bytes1, size1, bytes2, size2);

	/* common init */
	pSoundBuffer->SetCurrentPosition(0);
	audioSetVolume(Volume);

	PlayBuffer = FirstDecode = 1;
	hAccessNotification = CreateEvent(NULL, FALSE, FALSE, NULL);
	hWritingNotification = CreateEvent(NULL, FALSE, FALSE, NULL);

	hCtrlThread = CreateThread(NULL, 0, DSCtrlThread, NULL, 0, &tID);
	SetThreadPriority(hCtrlThread, THREAD_PRIORITY_HIGHEST);

	return (hCtrlThread != NULL);
}

void audioClose()
{
	PlayBuffer = 0;
	if (hAccessNotification) SetEvent(hAccessNotification);

	if (WaitForSingleObject(hCtrlThread, 3000) == WAIT_TIMEOUT) TerminateThread(hCtrlThread, -1);
	CloseHandle(hCtrlThread);

	if (hWritingNotification) {
		CloseHandle(hWritingNotification);
		hWritingNotification=NULL;
	}

	if (hAccessNotification) {
		CloseHandle(hAccessNotification);
		hAccessNotification=NULL;
	}

	if (pSoundBuffer) {
		pSoundBuffer->Stop();
		SAFE_RELEASE(pSoundBuffer);
	}

	if (pDS) SAFE_RELEASE(pDS);
}

int audioWrite(char* data,int len)
{
	void *bytes1,*bytes2;
	DWORD Size1,Size2,tID;

	RequestedInput = len;
	SetEvent(hAccessNotification);
	WaitForSingleObject(hWritingNotification, INFINITE);
	
	pSoundBuffer->Lock(WriteOffset, len, &bytes1, &Size1, &bytes2, &Size2, 0);
	if (!bytes2) CopyMemory(bytes1, data, len);
	else {
		CopyMemory(bytes1, data, Size1);
		CopyMemory(bytes2, data + Size1, Size2);
	}
	pSoundBuffer->Unlock(bytes1, Size1, bytes2, Size2);

	WriteOffset += len;
	WriteOffset %= BufferSize;

	if (FirstDecode) {
		FirstDecode = 0;
		pSoundBuffer->Play(0, 0, DSBPLAY_LOOPING);
	}

	return len;
}

void audioFlush()
{
}

/* linear variation */
void audioSetVolume(int vol)
{
	if (!pSoundBuffer) return;
	Volume = vol;
	pSoundBuffer->SetVolume((int)(1000.0f*Volume*(2.0f / 256.0f) - 2000));
}

int audioGetVolume()
{
	if (!pSoundBuffer) return 127;
	pSoundBuffer->GetVolume((LONG*)&Volume);
	Volume=(int)(((Volume/1000.0f) + 2) * 127);
	return Volume;
}