/*
 * Author : Ilyes Gouta
 * Subject : Win32 GUI code
 * Notes :
 *
 */

#include <windows.h>
#include <commctrl.h>
#include <dsound.h>
#include "resource.h"

#include "audio.h"
#include "dispatch.h"
#include "gui.h"
#include "vis.h"
#include "Version.h"

int get_frames(char* fn,int* count,float* duration);

#define DB      {                                              \
				char b[256];                                   \
				wsprintf(b,"Error code : %d",GetLastError());  \
				MessageBox(hWnd,b,App,MB_ICONSTOP);            \
				}                                              \

#define WM_EOF           WM_USER+1

extern char App[];

static HWND hWnd;
static HWND hProgress, hVolume, hStatic;
static HANDLE hPlayer;
static DWORD tID;

static char fileName[MAX_PATH];
static char playlistName[MAX_PATH];

static int playlistMode;
static int currentSong;
static int performNotify=1;
static int audioVolume;

static HANDLE hVisLib;
static char visName[MAX_PATH];
static winampVisModule* visModule;
static HANDLE hVisThread;
static int rendering_sw;

extern int decodedFrames;

static char notice[]="G-amp is based on an amp v0.7.6 port on Win32 by Gouta Ilyes 2002-2003.\r\nOriginal code (C) Tomislav Uzelac 1996-1997, modified since by many other people.";

void gui_init_playlist(HWND hWnd);
void gui_play_item(int index);
void gui_process_next();
void gui_open(HWND hWnd);

BOOL CALLBACK AboutProc(HWND hWnd,UINT Message,WPARAM wParam,LPARAM lParam)
{
	switch(Message) {
	case WM_INITDIALOG: {
		HRSRC res;HGLOBAL resource;char* data;
		res=FindResource(NULL,MAKEINTRESOURCE(ID_ORGLICENSE),"TEXT");
		if (!res) return TRUE;
		resource=LoadResource(NULL,res);
		data=(char*) LockResource(resource);
		SendDlgItemMessage(hWnd,ID_MINE,WM_SETTEXT,0,(LPARAM) notice);
		SendDlgItemMessage(hWnd,ID_ORIGINAL,WM_SETTEXT,0,(LPARAM) data);
		SetFocus(GetDlgItem(hWnd,IDCANCEL));
		return TRUE;
						}
	case WM_COMMAND:
	switch (LOWORD(wParam)) {
	case IDCANCEL:
		EndDialog(hWnd,TRUE);
		return TRUE;
	default: return FALSE;
	}
	default: return FALSE;
	}
}

BOOL CALLBACK PlaylistDlg(HWND hWnd,UINT Message,WPARAM wParam,LPARAM lParam)
{
	switch(Message) {
	case WM_INITDIALOG:
		gui_init_playlist(hWnd);
		return TRUE;
	case WM_COMMAND:
	switch (LOWORD(wParam)) {
	case IDCANCEL:
		EndDialog(hWnd,TRUE);
		return TRUE;
	}
	switch (HIWORD(wParam)) {
	case LBN_DBLCLK: {
		int index;
		index=SendDlgItemMessage(hWnd,ID_PLAYLISTFILES,LB_GETCURSEL,0,0);
		if (index==LB_ERR) return TRUE;
		gui_play_item(index);
		return TRUE;
					 }
	}
	default: return FALSE;
	}
}

DWORD WINAPI visualizationThread(LPVOID s)
{
	visModule->Init(visModule);
	while (rendering_sw) {
		MSG msg;
		visModule->Render(visModule);
		while (PeekMessage(&msg,NULL,0,0,PM_REMOVE)) DispatchMessage(&msg);
		Sleep(visModule->delayMs);
	}
	visModule->Quit(visModule);
	return 0;
}

DWORD WINAPI playerThread(LPVOID s)
{
	if (!fileName[0]) return 1;
	in_file=CreateFile(fileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if (in_file==INVALID_HANDLE_VALUE) return 1;

	EnableMenuItem(GetMenu(hWnd),ID_PLAY,MF_BYCOMMAND|MF_GRAYED);
	EnableMenuItem(GetMenu(hWnd),ID_STOP,MF_BYCOMMAND|MF_ENABLED);

	decodeMPEG();

	EnableMenuItem(GetMenu(hWnd),ID_PLAY,MF_BYCOMMAND|MF_ENABLED);
	EnableMenuItem(GetMenu(hWnd),ID_STOP,MF_BYCOMMAND|MF_GRAYED);
	
	close_audio();
	CloseHandle(in_file);

	if (performNotify) PostMessage(hWnd,WM_EOF,0,0); // we're done
	return 0;
}

int load_vis()
{
	winampVisGetHeaderType proc;
	winampVisHeader* visHeader;

	if (!visName[0]) return 0;
	hVisLib=LoadLibrary(visName);
	if (!hVisLib) return 0;

	proc=(winampVisGetHeaderType) GetProcAddress(hVisLib,"winampVisGetHeader");
	if (!proc) {
		FreeLibrary(hVisLib);hVisLib=NULL;
		ZeroMemory(visName,sizeof(visName));
		return 0;
	}

	visHeader=proc();
	if (!visHeader) {
		FreeLibrary(hVisLib);hVisLib=NULL;
		ZeroMemory(visName,sizeof(visName));
		return 0;
	}

	visModule=visHeader->getModule(0); // ask for the first module, generally it's enough
	if (!visModule) {
		FreeLibrary(hVisLib);hVisLib=NULL;
		ZeroMemory(visName,sizeof(visName));
		return 0;
	}

	visModule->hDllInstance=hVisLib;
	visModule->hwndParent=hWnd;
	visModule->sRate=44100;
	visModule->nCh=nch;
	visModule->waveformNch=1;
	visModule->spectrumNch=0; // ???
	return 1;
}

int start_vis()
{
	DWORD tID;

	if (!hVisLib) return 0;

	waveformData=visModule->waveformData;
	perform_vis=1;
	rendering_sw=1;

	hVisThread=CreateThread(NULL,0,visualizationThread,NULL,0,&tID);
	if (!hVisThread) return 0;

	return 1;
}

void free_vis()
{
	if (!hVisLib) return;
	perform_vis=0;
	rendering_sw=0;
	WaitForSingleObject(hVisThread,INFINITE);
	CloseHandle(hVisThread);hVisThread=NULL;
	FreeLibrary(hVisLib);hVisLib=NULL;
}

void config_vis()
{
	if (!hVisLib) return;
	visModule->Config(visModule);
}

void gui_init_playlist(HWND hWnd)
{
	FILE *playlist;int i=0;
	char buffer[MAX_PATH];
	char *p;

	if (!playlistMode) return;

	playlist=fopen(playlistName,"rt");
	if (!playlist) return;

	fgets(buffer,sizeof(buffer),playlist);
	fgets(buffer,sizeof(buffer),playlist);

	while (1) {
		for (p=buffer+sizeof(buffer);p>=buffer;p--) if (*p=='\n'||*p=='\r') *p=0; // fix a small fgets issue
		for (p=buffer;*p!=',';p++);
		SendDlgItemMessage(hWnd,ID_PLAYLISTFILES,LB_ADDSTRING,0,(LPARAM) ++p);
		fgets(buffer,sizeof(buffer),playlist); // skips the next two lines
		fgets(buffer,sizeof(buffer),playlist);
		if (feof(playlist)) break;
	}
	
	fclose(playlist);
	SendDlgItemMessage(hWnd,ID_PLAYLISTFILES,LB_SETCURSEL,currentSong-1,0);
}

void gui_play()
{
	float duration;int frames;
	char debug[256];
	get_frames(fileName,&frames,&duration);
	/*(int)duration/60,(int)duration%60*/
	SendMessage(hProgress,TBM_SETRANGEMIN,TRUE,0);
	SendMessage(hProgress,TBM_SETRANGEMAX,TRUE,frames);
	hPlayer=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) playerThread,NULL,0,&tID);
	SetTimer(hWnd,1,1000,NULL);
}

void gui_stop()
{
	if (hPlayer) {
		stopMPEG();
		if (WaitForSingleObject(hPlayer,INFINITE)==WAIT_TIMEOUT) {
			MessageBox(hWnd,"Failed asking thread to die !",App,MB_ICONSTOP); // will never happen cause INFINITE
			TerminateThread(hPlayer,0);
		}
		close_audio();
		if (in_file) {CloseHandle(in_file);in_file=NULL;}
		CloseHandle(hPlayer);
		hPlayer=NULL;
		KillTimer(hWnd,1);
	}

	EnableMenuItem(GetMenu(hWnd),ID_PLAY,MF_BYCOMMAND|MF_ENABLED);
	EnableMenuItem(GetMenu(hWnd),ID_STOP,MF_BYCOMMAND|MF_GRAYED);
}

void gui_play_item(int index)
{
	if (!playlistMode) return;
	performNotify=0;
	gui_stop();
	currentSong=index;
	gui_process_next();
	performNotify=1;
}

int gui_check_file(char* fn)
{
	FILE *f;
	f=fopen(fn,"rb");
	if (!f) return 0;
	fclose(f);
	return 1;
}

void gui_process_next()
{
	char buffer[MAX_PATH];
	char *p;int i=0;
	FILE *playlist;
	
	if (!playlistMode) return;

	playlist=fopen(playlistName,"rt");
	if (!playlist) return;

	fgets(buffer,sizeof(buffer),playlist); // skip the two first lines
	fgets(buffer,sizeof(buffer),playlist);

	fgets(buffer,sizeof(buffer),playlist);
	while (i<=currentSong) {
		for (p=buffer+sizeof(buffer);p>=buffer;p--) if (*p=='\n'||*p=='\r') *p=0;
		if (!gui_check_file(buffer)) {
			p=playlistName+lstrlen(playlistName);
			while (*p!='\\'&&p>=playlistName) p--;
			ZeroMemory(fileName,sizeof(fileName));
			lstrcpyn(fileName,playlistName,++p-playlistName+1);
			lstrcat(fileName,buffer);
		} else lstrcpy(fileName,buffer);
		fgets(buffer,sizeof(buffer),playlist);
		if (!fgets(buffer,sizeof(buffer),playlist)) {
			if (feof(playlist)) {currentSong=-1;break;}
		}
		i++;
	}
	
	currentSong++;
	fclose(playlist);

	gui_play();
}

void gui_about(HWND hWnd)
{
	DialogBox(GetModuleHandle(NULL),MAKEINTRESOURCE(ID_ABOUTDLG),hWnd,AboutProc);
}

int OnOpen(WPARAM wParam,LPARAM lParam)
{
	OPENFILENAME open;
	ZeroMemory(&open,sizeof(open));
	open.lStructSize=sizeof(open);
	open.hwndOwner=hWnd;
	open.hInstance=GetModuleHandle(NULL);
	open.lpstrFile=fileName;
	open.nMaxFile=sizeof(fileName);
	open.lpstrFilter="MPEG audio files (*.mp2,*.mp3)\0*.mp2;*.mp3\0";
	open.lpstrInitialDir=NULL;
	open.Flags=OFN_FILEMUSTEXIST|OFN_NONETWORKBUTTON|OFN_HIDEREADONLY;
	if (!GetOpenFileName(&open)) ZeroMemory(fileName,sizeof(fileName));
	playlistMode=0;
	return 0;
}

int OnVisSelect(WPARAM wParam,LPARAM lParam)
{
	OPENFILENAME open;
	ZeroMemory(&open,sizeof(open));
	open.lStructSize=sizeof(open);
	open.hwndOwner=hWnd;
	open.hInstance=GetModuleHandle(NULL);
	open.lpstrFile=visName;
	open.nMaxFile=sizeof(visName);
	open.lpstrFilter="Winamp visualization (*.dll)\0*.dll\0";
	open.lpstrInitialDir=NULL;
	open.Flags=OFN_FILEMUSTEXIST|OFN_NONETWORKBUTTON|OFN_HIDEREADONLY;
	if (!GetOpenFileName(&open)) ZeroMemory(visName,sizeof(visName));
	rendering_sw=0;
	load_vis();
	return 0;
}

int OnVisStop(WPARAM wParam,LPARAM lParam)
{
	free_vis();
	return 0;
}

int OnVisStart(WPARAM wParam,LPARAM lParam)
{
	start_vis();
	return 0;
}

int OnVisConf(WPARAM wParam,LPARAM lParam)
{
	config_vis();
	return 0;
}

int OnDestroy(WPARAM wParam,LPARAM lParam)
{
	gui_stop();
	PostQuitMessage(0);
	return 0;
}

int OnEof(WPARAM wParam,LPARAM lParam)
{
	gui_stop();
	if (playlistMode) gui_process_next();
	return 0;
}

int OnExit(WPARAM wParam,LPARAM lParam)
{
	DestroyWindow(hWnd);
	return 0;
}

int OnPlaylist(WPARAM wParam,LPARAM lParam)
{
	OPENFILENAME open;
	ZeroMemory(&open,sizeof(open));
	open.lStructSize=sizeof(open);
	open.hwndOwner=hWnd;
	open.hInstance=GetModuleHandle(NULL);
	open.lpstrFile=playlistName;
	open.nMaxFile=sizeof(playlistName);
	open.lpstrFilter="m3u playlist (*.m3u)\0*.m3u\0";
	open.lpstrInitialDir=NULL;
	open.Flags=OFN_FILEMUSTEXIST|OFN_NONETWORKBUTTON|OFN_HIDEREADONLY;
	if (!GetOpenFileName(&open)) return 0;
	if (hPlayer) gui_stop();
	playlistMode=1;currentSong=0;
	gui_process_next();
	return 0;
}

int OnAbout(WPARAM wParam,LPARAM lParam)
{
	gui_about(hWnd);
	return 0;
}

int OnPlay(WPARAM wParam,LPARAM lParam)
{
	gui_play();
	return 0;
}

int OnStop(WPARAM wParam,LPARAM lParam)
{
	playlistMode=0;
	gui_stop();
	return 0;
}

int OnShowPlaylist(WPARAM wParam,LPARAM lParam)
{
	DialogBox(GetModuleHandle(NULL),MAKEINTRESOURCE(ID_PLAYLISTDLG),hWnd,PlaylistDlg);
	return 0;
}

int OnKeys(WPARAM wParam,LPARAM lParam)
{
	if (wParam==VK_UP) audioVolume+=5;
	if (wParam==VK_DOWN) audioVolume-=5;
	if (audioVolume>255) audioVolume=255;
	if (audioVolume<0) audioVolume=0;
	audioSetVolume(audioVolume);
	return 0;
}

int OnTimer(WPARAM wParam,LPARAM lParam)
{
	SendMessage(hProgress,TBM_SETPOS,TRUE,decodedFrames);
	return 0;
}

int OnHScroll(WPARAM wParam,LPARAM lParam)
{
	static int two;

	if ((HWND)lParam == hProgress) {
		if (LOWORD(wParam)==TB_THUMBPOSITION) {
			if (two) {two=0;return 0;} else two=1;
			if (HIWORD(wParam)==decodedFrames) return 0;
			if (HIWORD(wParam)<decodedFrames) mpeg_rewind(decodedFrames-HIWORD(wParam));
			else mpeg_forward(HIWORD(wParam)-decodedFrames);
		}
	}

	if ((HWND)lParam == hVolume) {
		audioSetVolume(SendMessage(hVolume,TBM_GETPOS,0,0));
	}

	return 0;
}

int OnCommand(WPARAM wParam,LPARAM lParam)
{
	START_DISPATCH(LOWORD(wParam));
	DISPATCH(ID_EXIT,OnExit);
	DISPATCH(ID_OPEN,OnOpen);
	DISPATCH(ID_PLAYLIST,OnPlaylist);
	DISPATCH(ID_ABOUT,OnAbout);
	DISPATCH(ID_PLAY,OnPlay);
	DISPATCH(ID_STOP,OnStop);
	DISPATCH(ID_SHOWPLAYLIST,OnShowPlaylist);
	DISPATCH(ID_VISSELECT,OnVisSelect);
	DISPATCH(ID_VISSTOP,OnVisStop);
	DISPATCH(ID_VISSTART,OnVisStart);
	DISPATCH(ID_VISCONFIG,OnVisConf);
	END_COM_DISPATCH;
}

LRESULT CALLBACK WndProc(HWND hWnd,UINT Message,WPARAM wParam,LPARAM lParam)
{
	START_DISPATCH(Message);
	DISPATCH(WM_DESTROY,OnDestroy);
	DISPATCH(WM_EOF,OnEof);
	DISPATCH(WM_COMMAND,OnCommand);
	DISPATCH(WM_KEYDOWN,OnKeys);
	DISPATCH(WM_TIMER,OnTimer);
	DISPATCH(WM_HSCROLL,OnHScroll);
	END_DISPATCH;
}

int init_gui()
{
	WNDCLASS wndClass;
	char tmp[256];

	InitCommonControls();

	wndClass.style=CS_SAVEBITS|CS_BYTEALIGNCLIENT|CS_BYTEALIGNWINDOW|CS_HREDRAW|CS_VREDRAW;
	wndClass.lpfnWndProc=WndProc;
	wndClass.cbClsExtra=0;
	wndClass.cbWndExtra=0;
	wndClass.hInstance=GetModuleHandle(NULL);
	wndClass.hIcon=LoadIcon(wndClass.hInstance,MAKEINTRESOURCE(ID_ICON));
	wndClass.hCursor=LoadCursor(NULL,IDC_ARROW);
	wndClass.hbrBackground = (HBRUSH) (COLOR_BTNFACE+1);
	wndClass.lpszMenuName=MAKEINTRESOURCE(ID_MENU);
	wndClass.lpszClassName=App;

	wsprintf(tmp,"%s v%d.%d.%d",App,VERSION_MAJOR,VERSION_MINOR,VERSION_RELEASE);
	if (!RegisterClass(&wndClass)) return 0;
	hWnd=CreateWindow(App,tmp,WS_SYSMENU|WS_CAPTION|WS_MINIMIZEBOX,CW_USEDEFAULT,CW_USEDEFAULT,226,GetSystemMetrics(SM_CYMENU)+GetSystemMetrics(SM_CYCAPTION)+4*GetSystemMetrics(SM_CYBORDER)+43,NULL,NULL,GetModuleHandle(NULL),NULL);
	if (!hWnd) return 0;
	ShowWindow(hWnd,SW_NORMAL);

	hProgress=CreateWindow(TRACKBAR_CLASS,NULL,WS_CHILD|WS_VISIBLE|TBS_NOTICKS|TBS_HORZ|TBS_BOTH,0,0,220,20,hWnd,NULL,GetModuleHandle(NULL),NULL);
	hVolume=CreateWindow(TRACKBAR_CLASS,NULL,WS_CHILD|WS_VISIBLE|TBS_NOTICKS|TBS_HORZ|TBS_BOTH,48,22,172,18,hWnd,NULL,GetModuleHandle(NULL),NULL);
	hStatic=CreateWindow("STATIC",NULL,WS_CHILD|WS_VISIBLE,5,24,42,15,hWnd,NULL,GetModuleHandle(NULL),NULL);
	SendMessage(hStatic, WM_SETFONT, (WPARAM) GetStockObject(DEFAULT_GUI_FONT), 0);

	SendMessage(hVolume, TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(0,255));
	SendMessage(hVolume, TBM_SETPOS, TRUE, (LPARAM) audioGetVolume());
	SetWindowText(hStatic, "Volume");

	EnableMenuItem(GetMenu(hWnd),ID_PLAY,MF_BYCOMMAND|MF_ENABLED);
	EnableMenuItem(GetMenu(hWnd),ID_STOP,MF_BYCOMMAND|MF_GRAYED);

	ZeroMemory(playlistName,sizeof(playlistName));
	ZeroMemory(fileName,sizeof(fileName));
	ZeroMemory(visName,sizeof(visName));

	playlistMode=0;
	audioVolume=audioGetVolume();
	return 1;
}
