/*
 * Author : Ilyes Gouta
 * Subject : some MPEG audio utils.
 * Notes :
 *
 */

#include <windows.h>
#include <stdio.h>

typedef struct mp3header
{
	int IDex;
	int ID;
	int layer;
	int protection_bit;
	int bitrate_index;
	int sampling_frequency;
	int padding_bit;
	int private_bit;
	int mode;
	int mode_extension;
	int copyright;
	int original;
	int emphasis;
} mp3header;

#define BUFFER_SIZE       32
#define _SYNCWORD_          0x7ff

static unsigned char buffer[BUFFER_SIZE];
static unsigned int bufsize;
static unsigned int bitptr; // current bit ptr

static int mpeg1_bitrate_table[]={0,32,40,48,56,64,80,96,112,128,160,192,224,256,320,-1};
static int mpeg2_bitrate_table[]={0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,-1};
static int mpeg25_bitrate_table[]={0,8,16,24,32,40,48,56,64,-1};
static int* current_bitrate_table[]={mpeg25_bitrate_table,mpeg2_bitrate_table,NULL,mpeg1_bitrate_table};

static int mpeg1_sampling_table[]={44100,48000,32000};
static int mpeg2_sampling_table[]={22050,24000,16000};
static int mpeg25_sampling_table[]={11025,12000,8000};
static int* current_sampling_table[]={mpeg25_sampling_table,mpeg2_sampling_table,NULL,mpeg1_sampling_table};

static char* bitstream_table[]={"MPEG2.5","MPEG2",NULL,"MPEG1"};
static char* mode_table[]={"stereo","joint stereo","dual channel","single channel"};

static mp3header header;
#define INDEX ((header.ID<<1)|header.IDex)

static FILE* hFile;

void resetbits()
{
	bitptr=0;
	memset(buffer,0,sizeof(buffer));
}

static __inline int fillbuffer()
{
	return ((bufsize=fread(buffer,1,sizeof(buffer),hFile)));
}

static __inline unsigned int _getbits(int nbits) // 32 bits maximum.
{
	register unsigned int tmp;
	register unsigned int ptr;
	ptr = bitptr >> 3;
	tmp = (buffer[ptr]<<24) | (buffer[ptr+1]<<16) | (buffer[ptr+2]<<8) | (buffer[ptr+3]);
	tmp <<= bitptr & 7; // skips lower bits (used in the past)
	tmp >>= 32 - nbits; // masks the rest
	bitptr += nbits;
	return tmp;
}

static unsigned int mem_getbits(unsigned char* buf,int size,int nbits,int begin) // 32 bits maximum.
{
	register unsigned int tmp;
	register unsigned int ptr;
	static unsigned int loc;
	if (begin) loc=0; // restart from the beginning
	if ((loc+nbits)>8*size) return 0;
	ptr = loc >> 3;
	tmp = (buf[ptr]<<24) | (buf[ptr+1]<<16) | (buf[ptr+2]<<8) | (buf[ptr+3]);
	tmp <<= loc & 7; // skips lower bits (used in the past)
	tmp >>= 32 - nbits; // masks the rest
	loc += nbits;
	return tmp;
}

static int read_frameheader(unsigned char* buf)
{
#define GETBITS(a) mem_getbits(buf,4,a,0)
	int sword=mem_getbits(buf,4,11,1);
	if (sword!=_SYNCWORD_) return 0;
	header.IDex=GETBITS(1);
	header.ID=GETBITS(1);
	header.layer=GETBITS(2);
	header.protection_bit=GETBITS(1);
	header.bitrate_index=GETBITS(4);
	header.sampling_frequency=GETBITS(2);
	header.padding_bit=GETBITS(1);
	GETBITS(1);
	header.mode=GETBITS(2);
	return 1;
#undef GETBITS
}

static int _get_frames(float* duration)
{
	unsigned char frame_header[4];
	int count=0,framesize;
	char* data;
	
	data=(char*) malloc(1024);
	*duration=0;
	while (1) {
		if (!fread(frame_header,1,4,hFile)) break;
		if (!read_frameheader(frame_header)) break;
		framesize=(int)((header.ID==0?72:144)*((1000*current_bitrate_table[INDEX][header.bitrate_index])/(float)current_sampling_table[INDEX][header.sampling_frequency]));
		framesize+=header.padding_bit; // fixes frame size for the 44100, 22050, 11025 sampling rates
		*duration+=(float)framesize*8/(1000.f*current_bitrate_table[INDEX][header.bitrate_index]);
		fseek(hFile,framesize-4,SEEK_CUR);
		count++;
	}
	free(data);
	return (count);
}

int get_frames(char* fn,int* count,float* duration)
{
	char buf[3];
	
	hFile=fopen(fn,"rb");
	if (!hFile) return 0;
	
	fread(buf,3,1,hFile);
	if (buf[0]==0x49&&buf[1]==0x44&&buf[2]==0x33) { // ID3v2 tags
		char tag_size[4];DWORD tmp=0;
		fseek(hFile,3,SEEK_CUR);
		fread(tag_size,1,4,hFile);
		tmp=(((tag_size[0]<<24)|(tag_size[1]<<16)|(tag_size[2]<<8))>>1)|tag_size[3];
		fseek(hFile,tmp,SEEK_CUR);
	} else fseek(hFile,0,SEEK_SET);
	
	*count=_get_frames(duration);
	fclose(hFile);
	return 0;
}

int mpeg_forward(int count)
{
	return 0;
}

int mpeg_rewind(int count)
{
	return 0;
}